import { atom } from 'jotai';
import { QueryClient, QueryObserver, InfiniteQueryObserver, isCancelledError } from 'react-query';

const queryClientAtom = atom(new QueryClient());

function atomWithQuery(createQuery, getQueryClient = (get) => get(queryClientAtom)) {
  const queryDataAtom = atom((get) => {
    const queryClient = getQueryClient(get);
    const options = typeof createQuery === "function" ? createQuery(get) : createQuery;
    let settlePromise = null;
    const getInitialData = () => typeof options.initialData === "function" ? options.initialData() : options.initialData;
    const initialData = getInitialData();
    const dataAtom = atom(initialData || new Promise((resolve, reject) => {
      settlePromise = (data, err) => {
        if (err) {
          reject(err);
        } else {
          resolve(data);
        }
      };
    }));
    dataAtom.scope = queryAtom.scope;
    let setData = () => {
      throw new Error("atomWithQuery: setting data without mount");
    };
    const listener = (result) => {
      if (result.error) {
        if (settlePromise) {
          settlePromise(null, result.error);
          settlePromise = null;
        } else {
          setData(Promise.reject(result.error));
        }
        return;
      }
      if (result.data === void 0) {
        return;
      }
      if (settlePromise) {
        settlePromise(result.data);
        settlePromise = null;
      } else {
        setData(result.data);
      }
    };
    const defaultedOptions = queryClient.defaultQueryObserverOptions(options);
    if (typeof defaultedOptions.staleTime !== "number") {
      defaultedOptions.staleTime = 1e3;
    }
    const observer = new QueryObserver(queryClient, defaultedOptions);
    if (!initialData) {
      observer.fetchOptimistic(defaultedOptions).then(listener).catch((error) => listener({ error }));
    }
    dataAtom.onMount = (update) => {
      setData = update;
      const unsubscribe = observer.subscribe(listener);
      return unsubscribe;
    };
    return { dataAtom, observer };
  }, (get, set, action) => {
    switch (action.type) {
      case "refetch": {
        queryDataAtom.scope = queryAtom.scope;
        const { dataAtom, observer } = get(queryDataAtom);
        set(dataAtom, new Promise(() => {
        }));
        const p = observer.refetch({ cancelRefetch: true }).then(() => {
        });
        return p;
      }
    }
  });
  const queryAtom = atom((get) => {
    queryDataAtom.scope = queryAtom.scope;
    const { dataAtom } = get(queryDataAtom);
    return get(dataAtom);
  }, (_get, set, action) => {
    queryDataAtom.scope = queryAtom.scope;
    set(queryDataAtom, action);
  });
  return queryAtom;
}

function atomWithInfiniteQuery(createQuery, getQueryClient = (get) => get(queryClientAtom)) {
  const queryDataAtom = atom((get) => {
    const queryClient = getQueryClient(get);
    const options = typeof createQuery === "function" ? createQuery(get) : createQuery;
    let settlePromise = null;
    const getInitialData = () => typeof options.initialData === "function" ? options.initialData() : options.initialData;
    const initialData = getInitialData();
    const dataAtom = atom(initialData || new Promise((resolve, reject) => {
      settlePromise = (data, err) => {
        if (err) {
          reject(err);
        } else {
          resolve(data);
        }
      };
    }));
    dataAtom.scope = queryAtom.scope;
    let setData = () => {
      throw new Error("atomWithInfiniteQuery: setting data without mount");
    };
    const listener = (result) => {
      if (result.error && !isCancelledError(result.error)) {
        if (settlePromise) {
          settlePromise(null, result.error);
          settlePromise = null;
        } else {
          setData(Promise.reject(result.error));
        }
        return;
      }
      if (result.data === void 0) {
        return;
      }
      if (settlePromise) {
        settlePromise(result.data);
        settlePromise = null;
      } else {
        setData(result.data);
      }
    };
    const defaultedOptions = queryClient.defaultQueryObserverOptions(options);
    if (typeof defaultedOptions.staleTime !== "number") {
      defaultedOptions.staleTime = 1e3;
    }
    const observer = new InfiniteQueryObserver(queryClient, defaultedOptions);
    if (!initialData) {
      observer.fetchOptimistic(defaultedOptions).then(listener).catch((error) => listener({ error }));
    }
    dataAtom.onMount = (update) => {
      setData = update;
      const unsubscribe = observer == null ? void 0 : observer.subscribe(listener);
      return unsubscribe;
    };
    return { dataAtom, observer, options };
  }, (get, _set, action) => {
    queryDataAtom.scope = queryAtom.scope;
    const { observer } = get(queryDataAtom);
    switch (action.type) {
      case "refetch": {
        void observer.refetch();
        break;
      }
      case "fetchPreviousPage": {
        void observer.fetchPreviousPage();
        break;
      }
      case "fetchNextPage": {
        void observer.fetchNextPage();
        break;
      }
    }
  });
  const queryAtom = atom((get) => {
    queryDataAtom.scope = queryAtom.scope;
    const { dataAtom } = get(queryDataAtom);
    return get(dataAtom);
  }, (_get, set, action) => {
    queryDataAtom.scope = queryAtom.scope;
    set(queryDataAtom, action);
  });
  return queryAtom;
}

export { atomWithInfiniteQuery, atomWithQuery, queryClientAtom };
